# -*- coding: utf-8 -*-
"""SASExceptions

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1BADH17VoXYqdvpPNN0C15U0ycdGGKceQ
"""

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import object, range, map, zip
from io import open
'''
Created on Aug 2, 2010

@author: specuser

#******************************************************************************
# This file is part of RAW.
#
#    RAW is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    RAW is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with RAW.  If not, see <http://www.gnu.org/licenses/>.
#
#******************************************************************************
'''

# from __future__ import absolute_import, division, print_function, unicode_literals
# from builtins import object, range, map, zip
# from io import open

class ImageLoadError(Exception):
       def __init__(self, value):
           self.parameter = value
       def __str__(self):
           return repr(self.parameter)

class WrongImageFormat(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class MaskSizeError(Exception):
   def __init__(self, value):
       self.parameter = value
   def __str__(self):
       return repr(self.parameter)

class UnrecognizedDataFormat(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class DataNotCompatible(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class InvalidQrange(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class CenterNotFound(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class AbsScaleNormFailed(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class NormalizationError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class HeaderLoadError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class HeaderMaskLoadError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class NoATSASError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class HeaderSaveError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class GNOMError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class CorMapError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class EFAError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)